#pragma once

namespace JJX::json {
        enum class token_type { string, number, syntax, boolean, null };
        enum class value_type { string, number, object, array, boolean, null};
        struct token {
            std::string value;
            token_type type;
            int location;
            std::shared_ptr<std::string> full_source;
        };

        struct value
        {
            std::optional<std::string> string;
            std::optional<double> number;
            std::optional<bool> boolean;
            std::optional<std::vector<value>> array;
            std::optional<std::map<std::string, value>> object;
            value_type type;


            void operator=(double in);
            void operator=(const std::string& in);
            void operator=(bool in);
            void operator=(std::vector<value> in);
            void operator=(std::map<std::string, value> in);

            explicit operator double() const { return number.value(); }
            operator std::string() const { return string.value(); }
            explicit operator bool() const { return boolean.value(); }
            explicit operator std::vector<value>() const { return array.value(); }
            explicit operator std::map<std::string, value>() const { return object.value(); }


        };

        struct string_val : value
        {
            void operator=(const std::string& me);

            operator std::string();
        };
        struct number_val : value {};
        struct boolean_val : value {};
        struct object_val : value
        {
            void add(const std::string& key, value val);
            void add(const std::string& key, const std::string& val);
            value& operator[] (const std::string& key);

            template <typename T>
            T get(const std::string& key);

        };
        struct array_val : value
        {
            void add(value val);
            value& operator[] (int key);
        };

        std::tuple<std::vector<json::token>, std::string> lex(std::string);
        std::tuple<json::value, int, std::string> parse(std::vector<json::token>, int index = 0);
        std::tuple<json::value, std::string> parse(std::string);
        std::string deparse(json::value, std::string whitespace = "");

        string_val string(const std::string& text);

        number_val number(double input);

        boolean_val boolean(bool input);

        array_val array(std::vector<value> input);
        array_val array();

        object_val object(std::map<std::string, value> input);
        object_val object();

        template <typename T>
        value array_of(const std::vector<T>& input);
}
